/**
 * Playing a bit with PDF i fall of my chair when i
 * realized i have to do some complex stuff to handle
 * really the language...
 * ... so i take my book on compilator and begin to read
 * it...
 * After a while i decide to build 2 parser, one for the
 * pdf itself and the other for the data in the content
 * stream.
 *
 * These objects are gonna bigger when i implement more
 * stuff handling.
 */

const PDF_SYMBOLS = {
  Type: {length:1},
  Length: {length:1, length2:3},
  Contents: {length:1, length2:3,length:Math.E},
  Kids: {length:1,length2:3,lenght3:Math.E},
  Pages: {length:1,length2:3, length3:Math.E},
  Outlines: {length:1, length2:3, length3:Math.E},
  OpenAction: {length:1, length2:3, length3:Math.E},
  Annots: {length:1, length2:3, length3:Math.E},
  Count: {length:1},
  Parent: {length:3},
  Filter: {length:1},
  Subtype: {length:1}
};


const DATA_SYMBOLS = {
  "re": "re",
  "BT": "BT",
  "ET": "ET",
  "Tj": "Tj",
  "TJ": "TJ",
  "Tf": "Tf"
};

/**
 * PDF use the ASCII encoding for internal
 * purpose.
 * Thanks !
 *
 * Try to organize the lexem tables in order
 * to use the maximal munch rule
 * http://en.wikipedia.org/wiki/Maximal_munch
 *
 * We don't want ot use backtracing here, so keep simple
 *
 */
const PDF_LEXEM = [
  ["xref", /^xref$/],
  ["trailer", /^trailer$/],
  ["array_start", /^\[$/],
  ["array_end", /^\]$/],
  ["hexa", /<[a-f\d]+>/i],
  ["obj_end", /endobj/],
  ["obj_start", /obj/],
  ["stream_end", /endstream/],
  ["stream_start", /stream/],
  ["group_start", /^\<\<$/],
  ["group_end", /^\>\>$/],
  ["integer", /^[\d]+$/],
  ["float", /[\d]+\.[\d]+/],
  ["symbol", /\/[a-z]+/i],
  ["string", /[a-z]{2,}/i],
  ["char", /[a-z]/i]
];

/**
 * @name PDFParser
 * Base class for parsing PDF
 * This object contain lexical
 * and syntaxic analizer
 */
var PDFParser = {};


/**
 * @name PDFParser.lexical
 * Lexical parser for PDF, it
 * will performs basic operations
 * divide in two steps.
 * Firstly, it remove all separator
 * char from the string and clean up
 * some f***** data format like ^M
 * to have a better abstraction from the
 * system.
 * Secondly, it transform the stream
 * in lexem, giving back an array to the
 * syntaxic analizer.
 */
PDFParser.lexical = {};


/**
 * @name PDFParser.lexical._clean
 * @private
 *
 * Step 1 of the lexical analysis.
 * Clean up the string from all the
 * unused stuff.
 * @param {string} aString - The string to clean
 * @return {string} The cleaned string
 */
PDFParser.lexical._clean = function(aString) {
  aString = aString.replace(/^\s+/, "");

  //Add necessary space
  var elements = [/(<{2})/g, /(>{2})/g, /(\[)/g, /(\])/g];
  for each(element in elements) {
    aString = aString.replace(element, " $1 ");
  }
  elements = [/(\/)/g];
  for each(element in elements) {
    aString = aString.replace(element, " $1");
  }

  //Cut of unuseful \s
  aString = aString.replace(/\r?[\s]+/g, " ");
  aString = aString.replace(/\s$/, "");
  return aString;
};


/**
 * @name PDFParser.lexical._parse
 * Step 2 of the lexcial analysis.
 * Group the string in lexem and return
 * an array [ [type, data], ...]
 *
 * @param {string} aString - The string to parse
 * @return {array} The lexem array
 */
PDFParser.lexical._parse = function(aString) {
  aString = PDFParser.lexical._clean(aString);

  var lexems = [];
  var tokens = aString.split(" ");
  for each(token in tokens) {
    for (let i = 0; i < PDF_LEXEM.length; i++) {
      if (PDF_LEXEM[i][1].test(token)) {
        lexems.push([PDF_LEXEM[i][0], token]);
        break;
      }
    }
  }
  return lexems;
};


/**
 * @name PDFParser.syntaxic
 * Compute the data return by the lexical analysis
 * process and group the element in object if needed
 *
 */
PDFParser.syntaxic = {};


/**
 * @name PDFParser.syntaxic.exec
 *
 * Compute the data return by the lexical analysis
 * process and group the element in object if needed
 *
 * Non recognize symbols are simply returned as string.
 *
 * @param {string} The string to compute
 * @return {array} an array of object/string
 *
 */
PDFParser.syntaxic.exec = function(aString) {
  var lexems = PDFParser.lexical._parse( aString );
  var objects = [];

  //Now we have our lexems, so recompute the data
  //in a better way.
  for(let i = 0; i < lexems.length; i++) {
    switch (lexems[i][0]) {
      case "xref":
      case "trailer": //XXX do these special
      case "obj_start":
        objects.push({});
        break;

      case "symbol":
        this._fill(lexems, i, objects[objects.length-1]);
        //XXX Uncomment that
        //dump("----------------------" + objects[objects.length-1]["Contents"] + "\n");
        break;
    }
  }
};


/**
 * @name PDFParser.syntaxic._isIndirect
 * @private
 * Check if an object is an indirect ref or not
 *
 * @param {array} The lexems array
 * @param {integer} The entry in the lexems table
 * @return {boolean} Is the object an indirect ref
 *
 */
PDFParser.syntaxic._isIndirect = function(anArray, anIndex) {
  if(anArray[++anIndex][0] == "integer" &&
     anArray[++anIndex][0] == "integer" &&
     anArray[++anIndex][0] == "char")
      return true;

  return false;
};

/**
 * @name PDFParser.syntaxic._fill
 * @private
 *
 * Fill an object with data using the prefill
 * dictionnary
 * @param {array} The lexems array
 * @param {integer} The entry in the lexems table
 * @param {object} The object to fill
 */
PDFParser.syntaxic._fill = function(anArray, anIndex, anObject) {
  var symbol = anArray[anIndex][1].substr(1, anArray[anIndex][1].length - 1);
  var info = PDF_SYMBOLS[symbol];
  if (info) { //do we handle this
  //XXX do it!
    anObject[symbol] = info.length;
  }
};

